// zeroVM - VM crash course
// How to build a simple virtual machine
// Stefan Pietzonke 2019
//

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include <stdint.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>

typedef unsigned char           U1;		/* UBYTE   */
typedef int16_t                 S2;     /* INT     */
typedef uint16_t                U2;     /* UINT    */
typedef int32_t                 S4;     /* LONGINT */

typedef long long               S8;     /* 64 bit long */
typedef double                  F8;     /* DOUBLE */


#define DEBUG 0

#define EXE_NEXT(); ep = ep + eoffs; goto *jumpt[code[ep]];


/* opcodes */
#define PUSHB	0
#define PULLB	1

#define LOADB 	2
#define ADDI	3
#define SUBI 	4
#define MULI	5
#define DIVI 	6

#define INTR0	7


//	space for code and data
unsigned char code[1024];
unsigned char data[1024];


// new: function run: vm runtime
S2 run (void *arg)
{
	S8 ep;				// execution pointer
	S8 eoffs = 0;		// number of bytes to skip to get next opcode

	S8 regi[256];			// integer registers 64 bit
	S8 arg1, arg2, arg3;	// opcode arguments

	S8 cpu_core = 0;

	printf ("zeroVM\n");

	// data bytes 23 and 42
	data[0] = 23;
	data[1] = 42;

	// code
	// load byte from data_size
	code[0] = LOADB;
	code[1] = 0;	// data ptr
	code[2] = 0;	// offset
	code[3] = 0;	// target register

	// load byte from data
	code[4] = LOADB;
	code[5] = 1;
	code[6] = 0;
	code[7] = 1;

	// ADD
	code[8] = ADDI;
	code[9] = 0;	// register 0: 23
	code[10] = 1;	// register 1: 42
	code[11] = 2;	// target register 2: = 65

	code[12] = INTR0;
	code[13] = 4;	// interrupt 0: 4 = print number
	code[14] = 2;	// register 2: 65
	code[15] = 0;
	code[16] = 0;

	code[17] = INTR0;
	code[18] = 7;	// interrupt 0: 7 = print newline
	code[19] = 0;
	code[20] = 0;
	code[21] = 0;

	// MUL
	code[22] = MULI;
	code[23] = 0;	// register 0: 23
	code[24] = 1;	// register 1: 42
	code[25] = 3;	// target register 2: = 65

	code[26] = INTR0;
	code[27] = 4;	// interrupt 0: 4 = print number
	code[28] = 3;	// register 3: 966
	code[29] = 0;
	code[30] = 0;

	code[31] = INTR0;
	code[32] = 7;	// interrupt 0: 7 = print newline
	code[33] = 0;
	code[34] = 0;
	code[35] = 0;

	code[36] = INTR0;
	code[37] = 255;	// interrupt 0: 255 = EXIT
	code[38] = 2;
	code[39] = 0;
	code[40] = 0;

	// execution bytecode pointer
	ep = 0;

	// jumptable for indirect threading execution
	static void *jumpt[] =
    {
		&&pushb, &&pullb, &&loadb, &&addi, &&subi, &&muli, &&divi, &&intr0
	};

	printf ("running test program...\n\n");

	// fetch first bytecode
	EXE_NEXT();


	pushb:
	#if DEBUG
	printf ("%lli PUSHB\n", cpu_core);
	#endif
	arg1 = regi[code[ep + 1]];
	arg2 = regi[code[ep + 2]];
	arg3 = code[ep + 3];

	regi[arg3] = data[arg1 + arg2];

	eoffs = 4;
	EXE_NEXT();


	pullb:
	#if DEBUG
	printf ("%lli PULLB\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = regi[code[ep + 2]];
	arg3 = regi[code[ep + 3]];

	data[arg2 + arg3] = regi[arg1];

	eoffs = 4;
	EXE_NEXT();


	loadb:
	#if DEBUG
	printf ("%lli LOADB\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = data[arg1 + arg2];

	eoffs = 4;
	EXE_NEXT();


	addi:
	#if DEBUG
	printf ("%lli ADDI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] + regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	subi:
	#if DEBUG
	printf ("%lli SUBI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] - regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	muli:
	#if DEBUG
	printf ("%lli MULI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] * regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	divi:
	#if DEBUG
	printf ("%lli DIVI\n", cpu_core);
	#endif
	arg1 = code[ep + 1];
	arg2 = code[ep + 2];
	arg3 = code[ep + 3];

	regi[arg3] = regi[arg1] / regi[arg2];

	eoffs = 4;
	EXE_NEXT();


	intr0:
	arg1 = code[ep + 1];
	#if DEBUG
	printf ("%lli INTR0: %lli\n", cpu_core, arg1);
	#endif
	switch (arg1)
	{
		case 4:
			//printf ("PRINTI\n");
			arg2 = code[ep + 2];
			printf ("%lli", regi[arg2]);
			eoffs = 5;
			break;

		case 7:
			//printf ("PRINTNEWLINE\n");
			printf ("\n");
			eoffs = 5;
			break;

		case 8:
			//	DELAY
			arg2 = code[ep + 2];
			usleep (regi[arg2] * 1000);
			eoffs = 5;
			break;

		case 255:
			printf ("shutdown...\n");
			return (0);
	}
	EXE_NEXT();
}

int main (int ac, char *av[])
{
	run (NULL);
	exit (0);
}
